import { MessageTransmissionTransmitter } from "./msg.transmission.transmitter";
import { MessageTransmissionReceiver } from "./msg.transmission.receiver";
import { AdapterManager } from "../connector/adapter.manager";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionRequestResponse } from "./msg.transmission.request-response";
import { filter, Observable, Observer, Subject, Unsubscribable } from "rxjs";
import ConsoleLogger from "../utils/log.utils";
import { TransmitterAdapter } from "../connector/adapter.transmitter"
import { ReceiverAdapter } from "../connector/adapter.receiver"
import { MessageTransmissionManagerBase } from "../base/msg.transmission.manager.base";
import { ActorInterface, ActorProfile } from "../interface/actor.interface";
import { TransportService } from "../interface/interface";

export class MessageTransmissionManager<GeneralEvent> extends MessageTransmissionManagerBase<GeneralEvent> {
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionManager`, ['managers'])

    constructor(transportRef: Observable<TransportService>, browserEnv?: boolean) {
        super()
        this.console.log({ message: `Constructing self...` })
        transportRef.subscribe((transport: TransportService) => {
            this.outgoingBus.next({
                id: uuidv4(),
                type: `Transport Event`,
                event: `New Transport`,
                date: new Date(),
                data: transport
            })
        })
        if (browserEnv) this.browserEnv = browserEnv
        // Subscribe for adapterManager and it's relevent event
        this.adapterManager = new AdapterManager()
        this.adapterManager.subscribe(this.actorProfile, this.incomingBus, this.outgoingBus.asObservable)
    }

    public subscribe(actorProfile: ActorProfile, observer: Partial<Observer<GeneralEvent>>, subscribable?: ActorInterface<GeneralEvent>): Unsubscribable {
        const targetEvent: EventMessage = this.browserEnv ? 'New Server' : 'New Client';
        this.eventObj.transportEvent.pipe(
            filter(event => event.event == targetEvent)
        ).subscribe(event => {
            // get all adapters for all the connection
            let transmission: Transmission | undefined = this.instantiateTransmissionComponents(event?.data?.clientId)
            if (transmission) {
                observer.next({
                    id: uuidv4(),
                    type: `Transmission Event`,
                    event: 'New Transmission',
                    date: new Date(),
                    transmission: transmission
                })
            }
        })
    }

    protected instantiateTransmissionComponents(clientId: string): Transmission {
        let receiverInstance: MessageTransmissionReceiver = this.getReceiver(clientId, this.eventObj)
        let transmitterInstance: MessageTransmissionTransmitter = this.getTransmitter(clientId, this.eventObj)
        let requestResponseInstance: MessageTransmissionRequestResponse = this.getRequestResponse(clientId, this.eventObj, transmitterInstance, receiverInstance)
        let transmissionObj: Transmission = {
            clientId: clientId,
            transmitter: transmitterInstance,
            receiver: receiverInstance,
            requestResponse: requestResponseInstance,
            event: this.eventObj.globalEvent
        }

        return transmissionObj
    }

    private getReceiver(clientId: string, eventObj: EventObject): MessageTransmissionReceiver {
        throw new Error(`Method not defined`)
    }
    private getTransmitter(clientId: string, eventObj: EventObject): MessageTransmissionTransmitter {
        throw new Error(`Method not defined`)
    }
    private getRequestResponse(clientId: string, eventObj: EventObject, transmitter: MessageTransmissionTransmitter, receiver: MessageTransmissionReceiver): MessageTransmissionRequestResponse {
        throw new Error(`Method not defined`)
    }


}


